#! /usr/bin/env python
#
# Bot that hangs out in an IRC channel
# It parses messages and sends approriate links
# and such on to a MonkeyMan instance in Zope.
#
# Arguements: <URL of a Monkeyman instance>, <username>, <password>
#

import string
import re
from ircbot import SingleServerIRCBot
from irclib import nm_to_n, irc_lower

import xmlrpclib
import httplib
from base64 import encodestring

class BasicAuthTransport(xmlrpclib.Transport):
    """ Provides authentication for xml-rpc clients """
    def __init__(self, username=None, password=None, verbose=0):
        self.verbose=0
        self.username=username
        self.password=password

    def request(self, host, handler, request_body, verbose=0):
        # issue XML-RPC request

        h = httplib.HTTP(host)
        h.putrequest("POST", handler)

        # required by HTTP/1.1
        h.putheader("Host", host)

        # required by XML-RPC
        h.putheader("User-Agent", self.user_agent)
        h.putheader("Content-Type", "text/xml")
        h.putheader("Content-Length", str(len(request_body)))

        # basic auth
        if self.username is not None and self.password is not None:
            h.putheader("AUTHORIZATION", "Basic %s" % string.replace(
                encodestring("%s:%s" % (self.username, self.password)),
                "\012", ""))
        h.endheaders()

        if request_body:
            h.send(request_body)

        errcode, errmsg, headers = h.getreply()

        if errcode != 200:
            raise xmlrpclib.ProtocolError(
                host + handler,
                errcode, errmsg,
                headers
                )

        return self.parse_response(h.getfile())


class MonkeyBot(SingleServerIRCBot):
    def __init__(self, mm):
        self.mm = mm
        self.server, self.port, self.channel, self.nickname = self.mm.getConnectionInfo()
        SingleServerIRCBot.__init__(self, [(self.server, self.port)], self.nickname, self.nickname)
        self.ignores, self.ignore_prefixs, self.help_url = self.mm.getBehaviourInfo()
        self.m_alphacomment = re.compile(r'^([a-z])[:;,](.+)$')
        self.m_digitcomment = re.compile(r'^(\d+)[:;,](.+)$')
        self.m_titled_url = re.compile(r'^\s*(.*?)\s*[,:-]\s*(https?:\/\/\S+)')
        self.m_url = re.compile(r'^(.*?)(https?:\/\/\S+)')
        self.m_title = re.compile(r'^.*?@ITEM\s*(.+)$')
        self.m_set_title = re.compile(r'^\s*TITLE:?\s*(.*)$', re.IGNORECASE)
        self.m_set_url = re.compile(r'^\s*URL:?\s*(.*)$', re.IGNORECASE)
        self.m_start_junk = re.compile(r'^\s*\[')
        self.m_end_junk = re.compile(r'[\|,:;-]\s*$')
        
    def on_welcome(self, c, e):
        c.join(self.channel)

    def on_privmsg(self, c, e):
        #self.do_command(nm_to_n(e.source()), e.arguments()[0])
        self.on_pubmsg(self, c, e)

    def on_pubmsg(self, c, e):
        """
        this is where most of the parsing is done
        """
        
        line = e.arguments()[0]
        nick = nm_to_n(e.source())

        # check ignores
        if nick in self.ignores:
            return
        
        # check prefixs
        for pre in self.ignore_prefixs:
            if pre == line[0:len(pre)]:
                return

        # check for a item command referenced by letter
        if re.match(self.m_alphacomment,line):
            m = re.match(self.m_alphacomment,line)
            comment_ref, command = m.groups()
            self.parse_comment(nick, comment_ref, command)
        # check for a item command referenced by id
        elif re.match(self.m_digitcomment, line):
            m = re.match(self.m_digitcomment, line)
            comment_ref, command = m.groups()
            self.parse_comment(nick, comment_ref, command)
        # check for MonkeyItem postings with title and url
        elif re.match(self.m_titled_url, line):
            m = re.match(self.m_titled_url, line)
            title, url = m.groups()
            title = re.sub(self.m_start_junk,'',title) # remove start junk from the title
            title = re.sub(self.m_end_junk,'',title) # remove any trailing junk from the title
            self.post_item(nick,title,url)
        # check for a MonkeyItem posting with url only
        elif re.match(self.m_url, line):
            m = re.match(self.m_url, line)
            pre, url = m.groups()
            if pre in self.ignore_prefixs:
                return
            self.post_item(nick,'',url)
        # check for a MonkeyItem @ITEM title only posting
        elif re.match(self.m_title, line):
            m = re.match(self.m_title, line)
            title = m.groups()[0]
            msg = self.post_item(nick, title, '')
        # check for a command
        else:
            a = string.split(line, ":", 1)
            if len(a) > 1 and irc_lower(a[0]) == irc_lower(self.connection.get_nickname()):
                self.do_command(nm_to_n(e.source()), string.strip(a[1]))

    def post_item(self, nick, title, url):
        """
        post an item, making various checks before doing so
        and then informing the poster of the results
        """        
        # it's a go! post the item and let the user know
        msg = self.mm.addMonkeyItem(nick, title, url)
        self.connection.privmsg(self.channel, msg)
        
          
    def do_command(self, nick, cmd):
        c = self.connection
        
        if re.match(r'^\s*disconnect\s*$', cmd, re.I):
            self.disconnect()
        elif re.match(r'^\s*die\s*$', cmd, re.I):
            self.die()
        elif re.match(r'^\s*help\s*$', cmd, re.I):
            c.privmsg(self.channel,"See %s" % self.help_url)
        elif re.match(r'^\s*item:?\s*(.*)$', cmd, re.I):
            m = re.match(r'^\s*item:?\s*(.*)$', cmd, re.I)
            text = m.groups()[0]
            if len(text) < 3:
                # surely no one is going to try and post a two-letter item
                return
            self.post_item(nick,text,'')
        elif re.match(r'^\s*search\s+([\w\s]+?)\s*(\d+)?$', cmd, re.I):
            m = re.match(r'^\s*search\s+([\w\s]+?)\s*(\d+)?$', cmd, re.I)
            SearchableText, batch = m.groups()
            if not batch:
                batch = 5
            else:
                batch = int(batch)
            list = self.mm.searchMonkeyItemText( SearchableText, batch )
            if len(list) == 0:
                self.connection.notice(nick, "No matches were found.")
            else:
                for t in list:
                    self.connection.notice(nick, t)
        elif re.match(r'^\s*(view|last|list)\s*(\d+)?\s*$', cmd, re.I):
            m = re.match(r'^\s*(view|last|list)\s*(\d+)?\s*$', cmd, re.I)
            batch = m.groups()[1]
            list = self.mm.getLastMonkeyItems()
            if len(list) == 0:
                c.notice(nick, "I could not view the requested item. Perhaps *you* use the crack pipe?")
            else:
                for item in list:
                    c.notice(nick, item)

                
        elif cmd == 'stats':
            for chname, chobj in self.channels.items():
                c.notice(nick, "--- Channel statistics ---")
                c.notice(nick, "Channel: " + chname)
                users = chobj.users()
                users.sort()
                c.notice(nick, "Users: " + string.join(users, ", "))
                opers = chobj.opers()
                opers.sort()
                c.notice(nick, "Opers: " + string.join(opers, ", "))
                voiced = chobj.voiced()
                voiced.sort()
                c.notice(nick, "Voiced: " + string.join(voiced, ", "))
        else:
            c.notice(nick, "Not understood: " + cmd)

    def parse_comment( self, nick, comment_ref, command):
        """
        Quite often comments can contain commands that refer to the item
        they are commenting on
        """
        # if the command is empty, show the item
        if re.match(r'^\s*$', command):
            self.connection.notice(nick, self.mm.viewMonkeyItem(comment_ref) )
        # hide item
        if re.match(r'^\s*HIDE:?\s*$', command, re.IGNORECASE):
            self.mm.hideMonkeyItem(comment_ref)
        # show item
        elif re.match(r'^\s*SHOW:?\s*$', command, re.IGNORECASE):
            self.mm.showMonkeyItem(comment_ref)
        # set title of item
        elif re.match(self.m_set_title, command):
            m = re.match(self.m_set_title, command)
            title = m.groups()[0]
            self.mm.setMonkeyItemTitle(comment_ref, title)
        # set url of item
        elif re.match(self.m_set_url, command, re.IGNORECASE):
            m = re.match(self.m_set_url, command)
            url = m.groups()[0]
            self.mm.setMonkeyItemURL(comment_ref, url)
        # it must be a plain ol' comment, post it
        else:
            self.post_comment(nick, comment_ref, command)

    def post_comment(self, nick, comment_ref, line):
        """
        Add a comment to a posted item
        """
        # convert asterisks to italics
        line = re.sub(r'\*([^\*]+)\*',r'<i>\1</i>',line)

        #
        # regexs copied from DiaWebLog for use in the post_comment() method
        #
        # there may by some libs that can do a better job at parsing URLs and such than this
        #
        self.m_titlelink = re.compile(r'.*?\[?([^\[]+?)[\|,:-]\s*((https?://|ftp://|mailto:)[^\s\]]+)\]?')
        self.m_link = re.compile(r'.*?((https?:\/\/|ftp:\/\/)[^\s\]]+)')
        self.m_image = re.compile(r'.*?\+\[?((https?:\/\/|ftp:\/\/)[^\s\]]+)\]?')
        self.m_email = re.compile(r'.*?\[?([.\w]+)\@([.\w]+)\]?')
        
        if re.match(self.m_titlelink, line):
            # hyperlink with a title
            print re.match(self.m_titlelink, line).groups()
            title, link, protocol = re.match(self.m_titlelink, line).groups()
            line = '<a href="%s" id="monkeyman">%s</a>' % (link, title)
        elif re.match(self.m_link, line):
            # plain old link
            link, protocol = re.match(self.m_link, line).groups()
            line = '<a href="%s" id="monkeyman">%s</a>' % (link, link)            
        elif re.match(self.m_image, line):
            # inline image
            link, protocol = re.match(self.m_image, line).groups()
            line = '<span id="monkeyimage"><img src="%s" alt="" border="0" vspace="4" hspace="4" /></span>' % link
        elif re.match(self.m_email, line):
            # email spam protection
            left, right = re.match(self.m_email, line).groups()
            line = '<a href="mailto:%s&#64;%s" id="monkeyman">%s&#64;%s</a>' % (left ,right, left, right)
            
        self.mm.addComment(nick, comment_ref, '', line)

        
def main():    
    import sys
    if len(sys.argv) != 4:
        print "Usage: testbot <server> <username> <password>"
        sys.exit(1)

    trans =  BasicAuthTransport( sys.argv[2], sys.argv[3] )
    mm = xmlrpclib.Server(sys.argv[1],  transport = trans)
    bot = MonkeyBot(mm)
    bot.start()


if __name__ == "__main__":
    main()
